"""Client for the model registry."""
from __future__ import annotations

from collections.abc import Sequence

from ml_metadata.proto import MetadataStoreClientConfig

from .exceptions import StoreException
from .store import MLMDStore, ProtoType
from .types import ListOptions, ModelArtifact, ModelVersion, RegisteredModel
from .types.base import ProtoBase
from .types.options import MLMDListOptions


class ModelRegistryAPIClient:
    """Model registry API."""

    def __init__(
        self,
        server_address: str,
        port: int,
        client_key: str | None = None,
        server_cert: str | None = None,
        custom_ca: str | None = None,
    ):
        """Constructor.

        Args:
            server_address: Server address.
            port: Server port.
            client_key: The PEM-encoded private key as a byte string.
            server_cert: The PEM-encoded certificate as a byte string.
            custom_ca: The PEM-encoded root certificates as a byte string.
        """
        config = MetadataStoreClientConfig()
        config.host = server_address
        config.port = port
        if client_key is not None:
            config.ssl_config.client_key = client_key
        if server_cert is not None:
            config.ssl_config.server_cert = server_cert
        if custom_ca is not None:
            config.ssl_config.custom_ca = custom_ca
        self._store = MLMDStore(config)

    def _map(self, py_obj: ProtoBase) -> ProtoType:
        """Map a Python object to a proto object.

        Helper around the `map` method of the Python object.

        Args:
            py_obj: Python object.

        Returns:
            Proto object.
        """
        type_id = self._store.get_type_id(
            py_obj.get_proto_type(), py_obj.get_proto_type_name()
        )
        return py_obj.map(type_id)

    def upsert_registered_model(self, registered_model: RegisteredModel) -> str:
        """Upsert a registered model.

        Updates or creates a registered model on the server.
        This updates the registered_model instance passed in with new data from the servers.

        Args:
            registered_model: Registered model.

        Returns:
            ID of the registered model.
        """
        id = self._store.put_context(self._map(registered_model))
        new_py_rm = RegisteredModel.unmap(
            self._store.get_context(RegisteredModel.get_proto_type_name(), id)
        )
        id = str(id)
        registered_model.id = id
        registered_model.create_time_since_epoch = new_py_rm.create_time_since_epoch
        registered_model.last_update_time_since_epoch = (
            new_py_rm.last_update_time_since_epoch
        )
        return id

    def get_registered_model_by_id(self, id: str) -> RegisteredModel | None:
        """Fetch a registered model by its ID.

        Args:
            id: Registered model ID.

        Returns:
            Registered model.
        """
        proto_rm = self._store.get_context(
            RegisteredModel.get_proto_type_name(), id=int(id)
        )
        if proto_rm is not None:
            return RegisteredModel.unmap(proto_rm)

        return None

    def get_registered_model_by_params(
        self, name: str | None = None, external_id: str | None = None
    ) -> RegisteredModel | None:
        """Fetch a registered model by its name or external ID.

        Args:
            name: Registered model name.
            external_id: Registered model external ID.

        Returns:
            Registered model.
        """
        if name is None and external_id is None:
            msg = "Either name or external_id must be provided"
            raise StoreException(msg)
        proto_rm = self._store.get_context(
            RegisteredModel.get_proto_type_name(),
            name=name,
            external_id=external_id,
        )
        if proto_rm is not None:
            return RegisteredModel.unmap(proto_rm)

        return None

    def get_registered_models(
        self, options: ListOptions | None = None
    ) -> Sequence[RegisteredModel]:
        """Fetch registered models.

        Args:
            options: Options for listing registered models.

        Returns:
            Registered models.
        """
        mlmd_options = options.as_mlmd_list_options() if options else MLMDListOptions()
        proto_rms = self._store.get_contexts(
            RegisteredModel.get_proto_type_name(), mlmd_options
        )
        return [RegisteredModel.unmap(proto_rm) for proto_rm in proto_rms]

    def upsert_model_version(
        self, model_version: ModelVersion, registered_model_id: str
    ) -> str:
        """Upsert a model version.

        Updates or creates a model version on the server.
        This updates the model_version instance passed in with new data from the servers.

        Args:
            model_version: Model version to upsert.
            registered_model_id: ID of the registered model this version will be associated to.

        Returns:
            ID of the model version.
        """
        rm_id = int(registered_model_id)
        # this is not ideal but we need this info for the prefix
        model_version._registered_model_id = rm_id
        id = self._store.put_context(self._map(model_version))
        self._store.put_context_parent(rm_id, id)
        new_py_mv = ModelVersion.unmap(
            self._store.get_context(ModelVersion.get_proto_type_name(), id)
        )
        id = str(id)
        model_version.id = id
        model_version.create_time_since_epoch = new_py_mv.create_time_since_epoch
        model_version.last_update_time_since_epoch = (
            new_py_mv.last_update_time_since_epoch
        )
        return id

    def get_model_version_by_id(self, model_version_id: str) -> ModelVersion:
        """Fetch a model version by its ID.

        Args:
            model_version_id: Model version ID.

        Returns:
            Model version.
        """
        proto_mv = self._store.get_context(
            ModelVersion.get_proto_type_name(), id=int(model_version_id)
        )
        if proto_mv is not None:
            return ModelVersion.unmap(proto_mv)

        return None

    def get_model_versions(
        self, registered_model_id: str, options: ListOptions | None = None
    ) -> list[ModelVersion]:
        """Fetch model versions by registered model ID.

        Args:
            registered_model_id: Registered model ID.
            options: Options for listing model versions.

        Returns:
            Model versions.
        """
        mlmd_options = options.as_mlmd_list_options() if options else MLMDListOptions()
        mlmd_options.filter_query = f"parent_contexts_a.id = {registered_model_id}"
        return [
            ModelVersion.unmap(proto_mv)
            for proto_mv in self._store.get_contexts(
                ModelVersion.get_proto_type_name(), mlmd_options
            )
        ]

    def get_model_version_by_params(
        self,
        registered_model_id: str | None = None,
        version: str | None = None,
        external_id: str | None = None,
    ) -> ModelVersion | None:
        """Fetch a model version by associated parameters.

        Either fetches by using external ID or by using registered model ID and version.

        Args:
            registered_model_id: Registered model ID.
            version: Model version.
            external_id: Model version external ID.

        Returns:
            Model version.
        """
        if external_id is not None:
            proto_mv = self._store.get_context(
                ModelVersion.get_proto_type_name(), external_id=external_id
            )
        elif registered_model_id is None or version is None:
            msg = (
                "Either registered_model_id and version or external_id must be provided"
            )
            raise StoreException(msg)
        else:
            proto_mv = self._store.get_context(
                ModelVersion.get_proto_type_name(),
                name=f"{registered_model_id}:{version}",
            )
        if proto_mv is not None:
            return ModelVersion.unmap(proto_mv)

        return None

    def upsert_model_artifact(
        self, model_artifact: ModelArtifact, model_version_id: str
    ) -> str:
        """Upsert a model artifact.

        Updates or creates a model artifact on the server.
        This updates the model_artifact instance passed in with new data from the servers.

        Args:
            model_artifact: Model artifact to upsert.
            model_version_id: ID of the model version this artifact will be associated to.

        Returns:
            ID of the model artifact.
        """
        mv_id = int(model_version_id)
        try:
            self._store.get_attributed_artifact(
                ModelArtifact.get_proto_type_name(), mv_id
            )
            msg = f"Model version with ID {mv_id} already has a model artifact"
            raise StoreException(msg)
        except StoreException as e:
            if "found" not in str(e).lower():
                raise
        model_artifact._model_version_id = model_version_id
        id = self._store.put_artifact(self._map(model_artifact))
        self._store.put_attribution(mv_id, id)
        new_py_ma = ModelArtifact.unmap(
            self._store.get_artifact(ModelArtifact.get_proto_type_name(), id)
        )
        id = str(id)
        model_artifact.id = id
        model_artifact.create_time_since_epoch = new_py_ma.create_time_since_epoch
        model_artifact.last_update_time_since_epoch = (
            new_py_ma.last_update_time_since_epoch
        )
        return id

    def get_model_artifact_by_id(self, id: str) -> ModelArtifact | None:
        """Fetch a model artifact by its ID.

        Args:
            id: Model artifact ID.

        Returns:
            Model artifact.
        """
        proto_ma = self._store.get_artifact(
            ModelArtifact.get_proto_type_name(), int(id)
        )
        if proto_ma is not None:
            return ModelArtifact.unmap(proto_ma)

        return None

    def get_model_artifact_by_params(
        self, model_version_id: str | None = None, external_id: str | None = None
    ) -> ModelArtifact | None:
        """Fetch a model artifact either by external ID or by the ID of its associated model version.

        Args:
            model_version_id: ID of the associated model version.
            external_id: Model artifact external ID.

        Returns:
            Model artifact.
        """
        if external_id:
            proto_ma = self._store.get_artifact(
                ModelArtifact.get_proto_type_name(), external_id=external_id
            )
        elif not model_version_id:
            msg = "Either model_version_id or external_id must be provided"
            raise StoreException(msg)
        else:
            proto_ma = self._store.get_attributed_artifact(
                ModelArtifact.get_proto_type_name(), int(model_version_id)
            )
        if proto_ma is not None:
            return ModelArtifact.unmap(proto_ma)

        return None

    def get_model_artifacts(
        self,
        model_version_id: str | None = None,
        options: ListOptions | None = None,
    ) -> Sequence[ModelArtifact]:
        """Fetches model artifacts.

        Args:
            model_version_id: ID of the associated model version.
            options: Options for listing model artifacts.

        Returns:
            Model artifacts.
        """
        mlmd_options = options.as_mlmd_list_options() if options else MLMDListOptions()
        if model_version_id is not None:
            mlmd_options.filter_query = f"contexts_a.id = {model_version_id}"

        proto_mas = self._store.get_artifacts(
            ModelArtifact.get_proto_type_name(), mlmd_options
        )
        return [ModelArtifact.unmap(proto_ma) for proto_ma in proto_mas]



"""Exceptions for the model registry."""


class StoreException(Exception):
    """Storage related error."""


class UnsupportedTypeException(StoreException):
    """Raised when an unsupported type is encountered."""


class TypeNotFoundException(StoreException):
    """Raised when a type cannot be found."""


class ServerException(StoreException):
    """Raised when the server returns a bad response."""


class DuplicateException(StoreException):
    """Raised when the user tries to put an object with a conflicting property."""



"""MLMD storage backend wrapper."""

from __future__ import annotations

from collections.abc import Sequence
from typing import ClassVar

from ml_metadata import errors
from ml_metadata.metadata_store import ListOptions, MetadataStore
from ml_metadata.proto import (
    Artifact,
    Attribution,
    Context,
    MetadataStoreClientConfig,
    ParentContext,
)

from model_registry.exceptions import (
    DuplicateException,
    ServerException,
    StoreException,
    TypeNotFoundException,
    UnsupportedTypeException,
)

from .base import ProtoType


class MLMDStore:
    """MLMD storage backend."""

    # cache for MLMD type IDs
    _type_ids: ClassVar[dict[str, int]] = {}

    def __init__(self, config: MetadataStoreClientConfig):
        """Constructor.

        Args:
            config: MLMD config.
        """
        self._mlmd_store = MetadataStore(config)

    def get_type_id(self, mlmd_pt: type[ProtoType], type_name: str) -> int:
        """Get backend ID for a type.

        Args:
            mlmd_pt: Proto type.
            type_name: Name of the type.

        Returns:
            Backend ID.

        Raises:
            TypeNotFoundException: If the type doesn't exist.
            ServerException: If there was an error getting the type.
            UnsupportedTypeException: If the type is not supported.
        """
        if type_name in self._type_ids:
            return self._type_ids[type_name]

        if mlmd_pt is Artifact:
            mlmd_pt_name = "artifact"
            get_type = self._mlmd_store.get_artifact_type
        elif mlmd_pt is Context:
            mlmd_pt_name = "context"
            get_type = self._mlmd_store.get_context_type
        else:
            msg = f"Unsupported type: {mlmd_pt}"
            raise UnsupportedTypeException(msg)

        try:
            _type = get_type(type_name)
        except errors.NotFoundError as e:
            msg = f"{mlmd_pt_name} type {type_name} does not exist"
            raise TypeNotFoundException(msg) from e
        except errors.InternalError as e:
            msg = f"Couldn't get {mlmd_pt_name} type {type_name} from MLMD store"
            raise ServerException(msg) from e

        self._type_ids[type_name] = _type.id

        return _type.id

    def put_artifact(self, artifact: Artifact) -> int:
        """Put an artifact in the store.

        Args:
            artifact: Artifact to put.

        Returns:
            ID of the artifact.

        Raises:
            DuplicateException: If an artifact with the same name or external id already exists.
            TypeNotFoundException: If the type doesn't exist.
            StoreException: If the artifact isn't properly formed.
        """
        try:
            return self._mlmd_store.put_artifacts([artifact])[0]
        except errors.AlreadyExistsError as e:
            msg = f"Artifact {artifact.name} already exists"
            raise DuplicateException(msg) from e
        except errors.InvalidArgumentError as e:
            msg = "Artifact has invalid properties"
            raise StoreException(msg) from e
        except errors.NotFoundError as e:
            msg = f"Artifact type {artifact.type} does not exist"
            raise TypeNotFoundException(msg) from e

    def put_context(self, context: Context) -> int:
        """Put a context in the store.

        Args:
            context: Context to put.

        Returns:
            ID of the context.

        Raises:
            DuplicateException: If a context with the same name or external id already exists.
            TypeNotFoundException: If the type doesn't exist.
            StoreException: If the context isn't propertly formed.
        """
        try:
            return self._mlmd_store.put_contexts([context])[0]
        except errors.AlreadyExistsError as e:
            msg = f"Context {context.name} already exists"
            raise DuplicateException(msg) from e
        except errors.InvalidArgumentError as e:
            msg = "Context has invalid properties"
            raise StoreException(msg) from e
        except errors.NotFoundError as e:
            msg = f"Context type {context.type} does not exist"
            raise TypeNotFoundException(msg) from e

    def _filter_type(
        self, type_name: str, protos: Sequence[ProtoType]
    ) -> Sequence[ProtoType]:
        return [proto for proto in protos if proto.type == type_name]

    def get_context(
        self,
        ctx_type_name: str,
        id: int | None = None,
        name: str | None = None,
        external_id: str | None = None,
    ) -> Context | None:
        """Get a context from the store.

        This gets a context either by ID, name or external ID.
        If multiple arguments are provided, the simplest query will be performed.

        Args:
            ctx_type_name: Name of the context type.
            id: ID of the context.
            name: Name of the context.
            external_id: External ID of the context.

        Returns:
            Context.

        Raises:
            StoreException: Invalid arguments.
        """
        if name is not None:
            return self._mlmd_store.get_context_by_type_and_name(ctx_type_name, name)

        if id is not None:
            contexts = self._mlmd_store.get_contexts_by_id([id])
        elif external_id is not None:
            contexts = self._mlmd_store.get_contexts_by_external_ids([external_id])
        else:
            msg = "Either id, name or external_id must be provided"
            raise StoreException(msg)

        contexts = self._filter_type(ctx_type_name, contexts)
        if contexts:
            return contexts[0]

        return None

    def get_contexts(
        self, ctx_type_name: str, options: ListOptions
    ) -> Sequence[Context]:
        """Get contexts from the store.

        Args:
            ctx_type_name: Name of the context type.
            options: List options.

        Returns:
            Contexts.
        """
        # TODO: should we make options optional?
        # if options is not None:
        try:
            contexts = self._mlmd_store.get_contexts(options)
        except errors.InvalidArgumentError as e:
            msg = f"Invalid arguments for get_contexts: {e}"
            raise StoreException(msg) from e
        except errors.InternalError as e:
            msg = "Couldn't get contexts from MLMD store"
            raise ServerException(msg) from e

        contexts = self._filter_type(ctx_type_name, contexts)
        # else:
        #     contexts = self._mlmd_store.get_contexts_by_type(ctx_type_name)

        if not contexts:
            msg = f"Context type {ctx_type_name} does not exist"
            raise StoreException(msg)

        return contexts

    def put_context_parent(self, parent_id: int, child_id: int):
        """Put a parent-child relationship between two contexts.

        Args:
            parent_id: ID of the parent context.
            child_id: ID of the child context.

        Raises:
            StoreException: If the parent context doesn't exist.
            ServerException: If there was an error putting the parent context.
        """
        try:
            self._mlmd_store.put_parent_contexts(
                [ParentContext(parent_id=parent_id, child_id=child_id)]
            )
        except errors.AlreadyExistsError as e:
            msg = f"Parent context {parent_id} already exists for context {child_id}"
            raise StoreException(msg) from e
        except errors.InternalError as e:
            msg = f"Couldn't put parent context {parent_id} for context {child_id}"
            raise ServerException(msg) from e

    def put_attribution(self, context_id: int, artifact_id: int):
        """Put an attribution relationship between a context and an artifact.

        Args:
            context_id: ID of the context.
            artifact_id: ID of the artifact.

        Raises:
            StoreException: Invalid argument.
        """
        attribution = Attribution(context_id=context_id, artifact_id=artifact_id)
        try:
            self._mlmd_store.put_attributions_and_associations([attribution], [])
        except errors.InvalidArgumentError as e:
            if "artifact" in str(e).lower():
                msg = f"Artifact with ID {artifact_id} does not exist"
                raise StoreException(msg) from e

            if "context" in str(e).lower():
                msg = f"Context with ID {context_id} does not exist"
                raise StoreException(msg) from e

            msg = f"Invalid argument: {e}"
            raise StoreException(msg) from e

    def get_artifact(
        self,
        art_type_name: str,
        id: int | None = None,
        name: str | None = None,
        external_id: str | None = None,
    ) -> Artifact | None:
        """Get an artifact from the store.

        Gets an artifact either by ID, name or external ID.

        Args:
            art_type_name: Name of the artifact type.
            id: ID of the artifact.
            name: Name of the artifact.
            external_id: External ID of the artifact.

        Returns:
            Artifact.

        Raises:
            StoreException: Invalid arguments.
        """
        if name is not None:
            return self._mlmd_store.get_artifact_by_type_and_name(art_type_name, name)

        if id is not None:
            artifacts = self._mlmd_store.get_artifacts_by_id([id])
        elif external_id is not None:
            artifacts = self._mlmd_store.get_artifacts_by_external_ids([external_id])
        else:
            msg = "Either id, name or external_id must be provided"
            raise StoreException(msg)

        artifacts = self._filter_type(art_type_name, artifacts)
        if artifacts:
            return artifacts[0]

        return None

    def get_attributed_artifact(self, art_type_name: str, ctx_id: int) -> Artifact:
        """Get an artifact from the store by its attributed context.

        Args:
            art_type_name: Name of the artifact type.
            ctx_id: ID of the context.

        Returns:
            Artifact.
        """
        try:
            artifacts = self._mlmd_store.get_artifacts_by_context(ctx_id)
        except errors.InternalError as e:
            msg = f"Couldn't get artifacts by context {ctx_id}"
            raise ServerException(msg) from e
        artifacts = self._filter_type(art_type_name, artifacts)
        if artifacts:
            return artifacts[0]
        msg = "No artifacts found"
        raise StoreException(msg)

    def get_artifacts(
        self, art_type_name: str, options: ListOptions
    ) -> Sequence[Artifact]:
        """Get artifacts from the store.

        Args:
            art_type_name: Name of the artifact type.
            options: List options.

        Returns:
            Artifacts.
        """
        try:
            artifacts = self._mlmd_store.get_artifacts(options)
        except errors.InvalidArgumentError as e:
            msg = f"Invalid arguments for get_artifacts: {e}"
            raise StoreException(msg) from e
        except errors.InternalError as e:
            msg = "Couldn't get artifacts from MLMD store"
            raise ServerException(msg) from e

        artifacts = self._filter_type(art_type_name, artifacts)
        if not artifacts:
            msg = f"Artifact type {art_type_name} does not exist"
            raise StoreException(msg)

        return artifacts



"""Standard client for the model registry."""
from __future__ import annotations

from .core import ModelRegistryAPIClient
from .exceptions import StoreException
from .types import ModelArtifact, ModelVersion, RegisteredModel


class ModelRegistry:
    """Model registry client."""

    def __init__(
        self,
        server_address: str,
        port: int,
        author: str,
        client_key: str | None = None,
        server_cert: str | None = None,
        custom_ca: str | None = None,
    ):
        """Constructor.

        Args:
            server_address (str): Server address.
            port (int): Server port.
            author (str): Name of the author.
            client_key (str, optional): The PEM-encoded private key as a byte string.
            server_cert (str, optional): The PEM-encoded certificate as a byte string.
            custom_ca (str, optional): The PEM-encoded root certificates as a byte string.
        """
        # TODO: get args from env
        self._author = author
        self._api = ModelRegistryAPIClient(
            server_address, port, client_key, server_cert, custom_ca
        )

    def _register_model(self, name: str) -> RegisteredModel:
        if rm := self._api.get_registered_model_by_params(name):
            return rm

        rm = RegisteredModel(name)
        self._api.upsert_registered_model(rm)
        return rm

    def _register_new_version(
        self, rm: RegisteredModel, version: str, /, description: str | None
    ) -> ModelVersion:
        assert rm.id is not None, "Registered model must have an ID"
        if self._api.get_model_version_by_params(rm.id, version):
            msg = f"Version {version} already exists"
            raise StoreException(msg)

        mv = ModelVersion(rm.name, version, self._author, description=description)
        self._api.upsert_model_version(mv, rm.id)
        return mv

    def _register_model_artifact(
        self, mv: ModelVersion, uri: str, /, **kwargs
    ) -> ModelArtifact:
        assert mv.id is not None, "Model version must have an ID"
        ma = ModelArtifact(mv.model_name, uri, **kwargs)
        self._api.upsert_model_artifact(ma, mv.id)
        return ma

    def register_model(
        self,
        name: str,
        uri: str,
        *,
        model_format_name: str,
        model_format_version: str,
        version: str,
        description: str | None = None,
        storage_key: str | None = None,
        storage_path: str | None = None,
        service_account_name: str | None = None,
    ) -> RegisteredModel:
        """Register a model.

        Version has to be unique for the model.
        Either `storage_key` and `storage_path`, or `service_account_name` must be provided.

        Args:
            name (str): Name of the model.
            uri (str): URI of the model.

        Keyword Args:
            model_format_name (str): Name of the model format.
            model_format_version (str): Version of the model format.
            version (str): Version of the model.
            description (str, optional): Description of the model.
            storage_key (str, optional): Storage key.
            storage_path (str, optional): Storage path.
            service_account_name (str, optional): Service account name.

        Returns:
            RegisteredModel: Registered model.
        """
        rm = self._register_model(name)
        mv = self._register_new_version(rm, version, description=description)
        self._register_model_artifact(
            mv,
            uri,
            model_format_name=model_format_name,
            model_format_version=model_format_version,
            storage_key=storage_key,
            storage_path=storage_path,
            service_account_name=service_account_name,
        )

        return rm

    def get_registered_model(self, name: str) -> RegisteredModel:
        """Get a registered model.

        Args:
            name (str): Name of the model.

        Returns:
            RegisteredModel: Registered model.
        """
        return self._api.get_registered_model_by_params(name)

    def get_model_version(self, name: str, version: str) -> ModelVersion:
        """Get a model version.

        Args:
            name (str): Name of the model.
            version (str): Version of the model.

        Returns:
            ModelVersion: Model version.
        """
        rm = self._api.get_registered_model_by_params(name)
        return self._api.get_model_version_by_params(rm.id, version)

    def get_model_artifact(self, name: str, version: str) -> ModelArtifact:
        """Get a model artifact.

        Args:
            name (str): Name of the model.
            version (str): Version of the model.

        Returns:
            ModelArtifact: Model artifact.
        """
        mv = self.get_model_version(name, version)
        return self._api.get_model_artifact_by_params(mv.id)



"""Options for listing objects.

Provides a thin wrappers around the options classes defined in the MLMD Py lib.
"""

from __future__ import annotations

from enum import Enum

from attrs import define, field
from ml_metadata.metadata_store import ListOptions as MLMDListOptions
from ml_metadata.metadata_store import OrderByField as MLMDOrderByField


class OrderByField(Enum):
    """Fields to order by."""

    CREATE_TIME = MLMDOrderByField.CREATE_TIME
    UPDATE_TIME = MLMDOrderByField.UPDATE_TIME
    ID = MLMDOrderByField.ID


@define
class ListOptions:
    """Options for listing objects.

    Attributes:
        limit: Maximum number of objects to return.
        order_by: Field to order by.
        is_asc: Whether to order in ascending order. Defaults to True.
    """

    limit: int | None = field(default=None)
    order_by: OrderByField | None = field(default=None)
    is_asc: bool = field(default=True)

    def as_mlmd_list_options(self) -> MLMDListOptions:
        """Convert to MLMD ListOptions.

        Returns:
            MLMD ListOptions.
        """
        return MLMDListOptions(
            limit=self.limit,
            order_by=OrderByField(self.order_by).value if self.order_by else None,
            is_asc=self.is_asc,
        )



"""Artifact types for model registry.

Artifacts represent pieces of data.
This could be datasets, models, metrics, or any other piece of data produced or consumed by an
execution, such as an experiment run.

Those types are used to map between proto types based on artifacts and Python objects.

Todo:
    * Move part of the description to API Reference docs (#120).
"""

from __future__ import annotations

from enum import Enum, unique
from uuid import uuid4

from attrs import define, field
from ml_metadata.proto import Artifact
from typing_extensions import override

from .base import Prefixable, ProtoBase


@unique
class ArtifactState(Enum):
    """State of an artifact."""

    UNKNOWN = Artifact.UNKNOWN
    PENDING = Artifact.PENDING
    LIVE = Artifact.LIVE
    MARKED_FOR_DELETION = Artifact.MARKED_FOR_DELETION
    DELETED = Artifact.DELETED
    ABANDONED = Artifact.ABANDONED
    REFERENCE = Artifact.REFERENCE


@define(slots=False)
class BaseArtifact(ProtoBase):
    """Abstract base class for all artifacts.

    Attributes:
        name: Name of the artifact.
        uri: URI of the artifact.
        state: State of the artifact.
    """

    name: str
    uri: str
    state: ArtifactState = field(init=False, default=ArtifactState.UNKNOWN)

    @classmethod
    @override
    def get_proto_type(cls) -> type[Artifact]:
        return Artifact

    @override
    def map(self, type_id: int) -> Artifact:
        mlmd_obj = super().map(type_id)
        mlmd_obj.uri = self.uri
        mlmd_obj.state = self.state.value
        return mlmd_obj

    @classmethod
    @override
    def unmap(cls, mlmd_obj: Artifact) -> BaseArtifact:
        py_obj = super().unmap(mlmd_obj)
        assert isinstance(
            py_obj, BaseArtifact
        ), f"Expected BaseArtifact, got {type(py_obj)}"
        py_obj.uri = mlmd_obj.uri
        py_obj.state = ArtifactState(mlmd_obj.state)
        return py_obj


@define(slots=False, auto_attribs=True)
class ModelArtifact(BaseArtifact, Prefixable):
    """Represents a Model.

    Attributes:
        name: Name of the model.
        uri: URI of the model.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        model_format_name: Name of the model format.
        model_format_version: Version of the model format.
        storage_key: Storage secret name.
        storage_path: Storage path of the model.
        service_account_name: Name of the service account with storage secret.
    """

    # TODO: this could be an enum of valid formats
    model_format_name: str | None = field(kw_only=True, default=None)
    model_format_version: str | None = field(kw_only=True, default=None)
    storage_key: str | None = field(kw_only=True, default=None)
    storage_path: str | None = field(kw_only=True, default=None)
    service_account_name: str | None = field(kw_only=True, default=None)

    _model_version_id: str | None = field(init=False, default=None)

    @property
    @override
    def mlmd_name_prefix(self) -> str:
        return self._model_version_id if self._model_version_id else uuid4().hex

    @override
    def map(self, type_id: int) -> Artifact:
        mlmd_obj = super().map(type_id)
        props = {
            "model_format_name": self.model_format_name,
            "model_format_version": self.model_format_version,
            "storage_key": self.storage_key,
            "storage_path": self.storage_path,
            "service_account_name": self.service_account_name,
        }
        self._map_props(props, mlmd_obj.properties)
        return mlmd_obj

    @override
    @classmethod
    def unmap(cls, mlmd_obj: Artifact) -> ModelArtifact:
        py_obj = super().unmap(mlmd_obj)
        assert isinstance(
            py_obj, ModelArtifact
        ), f"Expected ModelArtifact, got {type(py_obj)}"
        py_obj.model_format_name = mlmd_obj.properties["model_format_name"].string_value
        py_obj.model_format_version = mlmd_obj.properties[
            "model_format_version"
        ].string_value
        py_obj.storage_key = mlmd_obj.properties["storage_key"].string_value
        py_obj.storage_path = mlmd_obj.properties["storage_path"].string_value
        py_obj.service_account_name = mlmd_obj.properties[
            "service_account_name"
        ].string_value
        return py_obj



"""Base classes and types for MLMD store."""

from typing import Union

from ml_metadata.proto import Artifact, Context

ProtoType = Union[Artifact, Context]
"""Union of all proto types. """
ScalarType = Union[str, int, float, bool]
"""Union of all scalar types.

Those types are easy to map to and from proto types, and can also be queried in MLMD.
"""



"""Context types for model registry.

Contexts group related Artifacts together.
They provide a way to organize and categorize components in a workflow.

Those types are used to map between proto types based on contexts and Python objects.

Todo:
    * Move part of the description to API Reference docs (#120).
"""

from __future__ import annotations

from abc import ABC
from enum import Enum, unique

from attrs import define, field
from ml_metadata.proto import Context
from typing_extensions import override

from model_registry.store import ScalarType

from .artifacts import BaseArtifact
from .base import Prefixable, ProtoBase


@unique
class ContextState(Enum):
    """State of the context.

    LIVE: The context is live and can be used.
    ARCHIVED: The context is archived and can't be used.
    """

    LIVE = "LIVE"
    ARCHIVED = "ARCHIVED"


@define(slots=False, init=False)
class BaseContext(ProtoBase, ABC):
    """Abstract base class for all contexts."""

    state: ContextState = field(init=False, default=ContextState.LIVE)

    @override
    def map(self, type_id: int) -> Context:
        mlmd_obj = super().map(type_id)
        mlmd_obj.properties["state"].string_value = self.state.value
        return mlmd_obj

    @classmethod
    @override
    def unmap(cls, mlmd_obj: Context) -> BaseContext:
        py_obj = super().unmap(mlmd_obj)
        assert isinstance(
            py_obj, BaseContext
        ), f"Expected BaseContext, got {type(py_obj)}"
        py_obj.state = ContextState(mlmd_obj.properties["state"].string_value)
        return py_obj

    @classmethod
    @override
    def get_proto_type(cls) -> type[Context]:
        return Context


@define(slots=False)
class ModelVersion(BaseContext, Prefixable):
    """Represents a model version.

    Attributes:
        model_name: Name of the model associated with this version.
        version: Version of the model.
        author: Author of the model version.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        artifacts: Artifacts associated with this version.
        tags: Tags associated with this version.
        metadata: Metadata associated with this version.
    """

    model_name: str
    version: str
    author: str
    artifacts: list[BaseArtifact] = field(init=False, factory=list)
    tags: list[str] = field(init=False, factory=list)
    metadata: dict[str, ScalarType] = field(init=False, factory=dict)

    _registered_model_id: int | None = field(init=False, default=None)

    def __attrs_post_init__(self) -> None:
        self.name = self.version

    @property
    @override
    def mlmd_name_prefix(self):
        assert (
            self._registered_model_id is not None
        ), "There's no registered model associated with this version"
        return self._registered_model_id

    @override
    def map(self, type_id: int) -> Context:
        mlmd_obj = super().map(type_id)
        # this should match the name of the registered model
        mlmd_obj.properties["model_name"].string_value = self.model_name
        mlmd_obj.properties["author"].string_value = self.author
        if self.tags:
            mlmd_obj.properties["tags"].string_value = ",".join(self.tags)
        self._map_props(self.metadata, mlmd_obj.custom_properties)
        return mlmd_obj

    @classmethod
    @override
    def unmap(cls, mlmd_obj: Context) -> ModelVersion:
        py_obj = super().unmap(mlmd_obj)
        assert isinstance(
            py_obj, ModelVersion
        ), f"Expected ModelVersion, got {type(py_obj)}"
        py_obj.version = py_obj.name
        py_obj.model_name = mlmd_obj.properties["model_name"].string_value
        py_obj.author = mlmd_obj.properties["author"].string_value
        tags = mlmd_obj.properties["tags"].string_value
        if tags:
            tags = tags.split(",")
        py_obj.tags = tags or []
        py_obj.metadata = cls._unmap_props(mlmd_obj.custom_properties)
        return py_obj


@define(slots=False)
class RegisteredModel(BaseContext):
    """Represents a registered model.

    Attributes:
        name: Registered model name.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
    """

    name: str

    @classmethod
    @override
    def unmap(cls, mlmd_obj: Context) -> RegisteredModel:
        py_obj = super().unmap(mlmd_obj)
        assert isinstance(
            py_obj, RegisteredModel
        ), f"Expected RegisteredModel, got {type(py_obj)}"
        return py_obj



"""Model registry types.

Types are based on [ML Metadata](https://github.com/google/ml-metadata), with Pythonic class wrappers.
"""

from .artifacts import ArtifactState, ModelArtifact
from .contexts import ContextState, ModelVersion, RegisteredModel
from .options import ListOptions, OrderByField

__all__ = [
    # Artifacts
    "ModelArtifact",
    "ArtifactState",
    # Contexts
    "ModelVersion",
    "RegisteredModel",
    "ContextState",
    # Options
    "ListOptions",
    "OrderByField",
]



"""Base types for model registry."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define, field
from typing_extensions import override

from model_registry.store import ProtoType, ScalarType


class Mappable(ABC):
    """Interface for types that can be mapped to and from proto types."""

    @classmethod
    def get_proto_type_name(cls) -> str:
        """Name of the proto type.

        Returns:
            Name of the class prefixed with `odh.`
        """
        return f"odh.{cls.__name__}"

    @property
    @abstractmethod
    def proto_name(self) -> str:
        """Name of the proto object."""
        pass

    @abstractmethod
    def map(self, type_id: int) -> ProtoType:
        """Map to a proto object.

        Args:
            type_id (int): ID of the type.

        Returns:
            Proto object.
        """
        pass

    @classmethod
    @abstractmethod
    def unmap(cls, mlmd_obj: ProtoType) -> Mappable:
        """Map from a proto object.

        Args:
            mlmd_obj: Proto object.

        Returns:
            Python object.
        """
        pass


class Prefixable(ABC):
    """Interface for types that are prefixed.

    We use prefixes to ensure that the user can insert more than one instance of the same type
    with the same name/external_id.
    """

    @property
    @abstractmethod
    def mlmd_name_prefix(self) -> str:
        """Prefix to be used in the proto object."""
        pass


@define(slots=False, init=False)
class ProtoBase(Mappable, ABC):
    """Abstract base type for protos.

    This is a type defining common functionality for all types representing Model Registry protos,
    such as Artifacts, Contexts, and Executions.

    Attributes:
        id: Protobuf object ID. Auto-assigned when put on the server.
        name: Name of the object.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        create_time_since_epoch: Seconds elapsed since object creation time, measured against epoch.
        last_update_time_since_epoch: Seconds elapsed since object last update time, measured against epoch.
    """

    name: str = field(init=False)
    id: str | None = field(init=False, default=None)
    description: str | None = field(kw_only=True, default=None)
    external_id: str | None = field(kw_only=True, default=None)
    create_time_since_epoch: int | None = field(init=False, default=None)
    last_update_time_since_epoch: int | None = field(init=False, default=None)

    @property
    @override
    def proto_name(self) -> str:
        if isinstance(self, Prefixable):
            return f"{self.mlmd_name_prefix}:{self.name}"
        return self.name

    @classmethod
    @abstractmethod
    def get_proto_type(cls) -> type[ProtoType]:
        """Proto type associated with this class.

        Returns:
            Proto type.
        """
        pass

    @staticmethod
    def _map_props(
        py_props: Mapping[str, ScalarType | None], mlmd_props: dict[str, Any]
    ):
        """Map properties from Python to proto.

        Args:
            py_props: Python properties.
            mlmd_props: Proto properties, will be modified in place.
        """
        for key, value in py_props.items():
            if value is None:
                continue
            # TODO: use pattern matching here (3.10)
            if isinstance(value, bool):
                mlmd_props[key].bool_value = value
            elif isinstance(value, int):
                mlmd_props[key].int_value = value
            elif isinstance(value, float):
                mlmd_props[key].double_value = value
            elif isinstance(value, str):
                mlmd_props[key].string_value = value
            else:
                msg = f"Unsupported type: {type(value)}"
                raise Exception(msg)

    @override
    def map(self, type_id: int) -> ProtoType:
        mlmd_obj = (self.get_proto_type())()
        mlmd_obj.name = self.proto_name
        mlmd_obj.type_id = type_id
        if self.id:
            mlmd_obj.id = int(self.id)
        if self.external_id:
            mlmd_obj.external_id = self.external_id
        if self.description:
            mlmd_obj.properties["description"].string_value = self.description
        return mlmd_obj

    @staticmethod
    def _unmap_props(mlmd_props: dict[str, Any]) -> dict[str, ScalarType]:
        """Map properties from proto to Python.

        Args:
            mlmd_props: Proto properties.

        Returns:
            Python properties.
        """
        py_props: dict[str, ScalarType] = {}
        for key, value in mlmd_props.items():
            # TODO: use pattern matching here (3.10)
            if value.HasField("bool_value"):
                py_props[key] = value.bool_value
            elif value.HasField("int_value"):
                py_props[key] = value.int_value
            elif value.HasField("double_value"):
                py_props[key] = value.double_value
            elif value.HasField("string_value"):
                py_props[key] = value.string_value
            else:
                msg = f"Unsupported type: {type(value)}"
                raise Exception(msg)

        return py_props

    T = TypeVar("T", bound="ProtoBase")

    @classmethod
    @override
    def unmap(cls: type[T], mlmd_obj: ProtoType) -> T:
        py_obj = cls.__new__(cls)
        py_obj.id = str(mlmd_obj.id)
        if isinstance(py_obj, Prefixable):
            name: str = mlmd_obj.name
            assert ":" in name, f"Expected {name} to be prefixed"
            py_obj.name = name.split(":", 1)[1]
        else:
            py_obj.name = mlmd_obj.name
        py_obj.description = mlmd_obj.properties["description"].string_value
        py_obj.external_id = mlmd_obj.external_id
        py_obj.create_time_since_epoch = mlmd_obj.create_time_since_epoch
        py_obj.last_update_time_since_epoch = mlmd_obj.last_update_time_since_epoch
        return py_obj
